// 2 axes joysticks drives 2 servos :

// written by hobby coder - use at own risk, author accepts no responsibility for anything, anywhere, ever

// Status 16 September 2020
//   tested on EasyPIC 7 development board with seperate power supply for servos (common ground)
//   DOES NOT WORK if servo powered from development board - servos create too much noise on supply
//   Commenting for Timer 0 low seems a little off - too lazy to work out why - I did it all by trial and error with scope anyway
//   too lazy to comment timer 1 calculations - perhaps I will get bored one day and try it

// Be sure to power the Servo Motors independantly of the PIC - Noise from the Servo will make the PIC go crazy
// Servo motor control signals on RB0 and RB1
// Analogue inputs on RA0 and RA1
// assumes that standard hobby servos are being used

// MCU : PIC 16F887, Eternal 4 MHz crystal
// Driving of 2 servomotors :    - servomotor LOWER (360  turn) on PORTD.B0 (pin 19),
//            > Servomotor TOWER PRO MG995 (360 ),
//            > Timer 0,
//            - servomotor UPPER (180 /270  turn) on PORTD.B1 (pin 20),
//            > Servomotor TOWER PRO MG996R (180 /270 ),
//            > Timer 1.

// global variables
volatile unsigned char PWM1_level_next;                                         // a simple flag to signal when to go low or high next time
volatile unsigned char PWM2_level_next;                                         // a simple flag to signal when to go low or high next time
volatile unsigned int PWM1_high_time = 0;                                       // PWM1 time control - 0 to 252 gives 0.5 to 2.5 ms
volatile unsigned int PWM2_high_time = 0;                                       // PWM2 time control - 0 to 252 gives 0.5 to 2.5 ms
volatile unsigned char PWM1_low_count = 0;                                      // timer 0 has limited timing capability - so use it 5 and a bit times

void Interrupt(){
  unsigned int TMR1_value;                                                      // temporary calculation value

    if (TMR0IF_bit){                                                            // handle timer 0 interrupt
        TMR0IF_bit = 0;
        if(PWM1_level_next){                                                    // was high - now needs to go low
            PORTD.B0 = 1;                                                       // PWM1 on RB0 high 0.5 to 2.5 ms
            PWM1_low_count = 0;
            PWM1_level_next = 0;                                                // set next flag low
            TMR0 = 226 - (PWM1_high_time >> 1);                                 // set new high time. 0.5mS minimum, 2.5ms max
        }
        else{                                                                   // we are in low time
            PORTD.B0 = 0;                                                       // PWM1 on RB0 low
            PWM1_low_count++;                                                   // T0 times out 5 times to extend low time  5 x 3.25 = 16.25 ms
            if(PWM1_low_count < 6){                                             // count longish time x 6 to get over limited T0 timing capability
                TMR0 = 43;                                                      // set new 1 sixth of low time - 3.25 ms
            }
            else{                                                               // 16.25 ms done - now add on the last little bit of low time
                TMR0 = 114 + (PWM1_high_time >> 1);                             // set final adjusted low time 256-114 to 256-114+(252/2) - leaves 142 to 16
                PWM1_level_next = 1;                                            // set next flag high
                PWM1_low_count = 0;                                             // reset 5-count ready for next time
            }
        }
    }

    if (TMR1IF_bit){                                                            // handle timer 1 interrupt
        TMR1IF_bit = 0;
        if(PWM2_level_next){                                                    // was high - next time needs to go low
            PORTD.B1 = 1;                                                       // PWM on RB1 high
            PWM2_level_next = 0;                                                // set next flag low
            TMR1_value = 65479 - (PWM2_high_time);                              // calculate 16 bit timer 1 value adjusted for high time
            TMR1H = TMR1_value >> 8;                                            // high byte
            TMR1L = TMR1_value & 0xff;                                          // set new high time.
        }
        else{                                                                   // we are in low time
            PORTD.B1 = 0;                                                       // PWM2 on RB1 low
            TMR1_value = 63100 + (PWM2_high_time);                              // 16 bit timer 1 value adjusted for low time
            TMR1H = TMR1_value >> 8;                                            // high byte
            TMR1L = TMR1_value & 0xff;                                          // set new low time
            PWM2_level_next = 1;                                                // set flag for next time needs to go high
        }
    }
}

void main(){

    TRISC = 0;                                                                  // all output
    TRISD = 0;                                                                  // all output
    PORTC = 0;                                                                  // all outputs low
    PORTD = 0;                                                                  // all outputs low

    OPTION_REG = 0x83;                                                          // Timer0 Prescaler 1:4, FOSC/4, On

    T1CON = 0x31;                                                               // 1:8 prescale, FOSC/4, On
    TMR1IF_bit = 0;                                                             // disable T1 Interrupt while adjusting
    TMR1H = 0xff;                                                               // don't care - will be adjusted later
    TMR1L = 0xE0;                                                               // don't care - will be adjusted later
    TMR1IE_bit = 1;                                                             // now ready for interrupt
    INTCON = 0xE0;                                                              // enable T0 int, peripheral int, global int

    while(1){                                                                   // main loop
        PWM1_high_time = ADC_Read(0)/4;                                         // 0 = 0.5mS, 252 = 2.5mS
        PWM2_high_time = ADC_Read(1)/4;                                         // 0 = 0.5mS, 252 = 2.5mS
    }
}